﻿namespace FlickRSearcher.ViewModels
{
  using System;
  using System.Collections.ObjectModel;
  using System.ComponentModel;
  using System.Net;
  using System.Reactive.Linq;
  using System.Windows.Input;
  using FlickRSearcher.Utility;

  public enum MainViewVisualStates
  {
    SearchVisible,
    SearchCollapsed,
    NoResults,
    Results
  }
  public class MainViewModel : PropertyChangedNotification
  {  
    public MainViewModel()
    {
      this.Images = new ObservableCollection<ImageViewModel>();
      this.CurrentPage = 1;

      InitialseCommands();
      InitialiseSearchTextSubscription();      
    }
    /// <summary>
    /// This is a bit of a hack because it gives the impression that there
    /// is only one visual state when there isn't but it does what I need
    /// it to do here.
    /// </summary>
    public string VisualState
    {
      get
      {
        return (this.InternalVisualState.ToString());
      }
    }

    MainViewVisualStates InternalVisualState
    {
      get
      {
        return (this.internalVisualState);
      }
      set
      {
        this.internalVisualState = value;
        RaisePropertyChanged("VisualState");
      }
    }
    MainViewVisualStates internalVisualState;

    public ICommand CancelCommand
    {
      get
      {
        return (this.cancelCommand);
      }
    }
    SimpleCommand cancelCommand;

    public ICommand NextPageCommand
    {
      get
      {
        return (nextPageCommand);
      }
    }
    SimpleCommand nextPageCommand;

    public ICommand PreviousPageCommand
    {
      get
      {
        return (previousPageCommand);
      }
    }
    SimpleCommand previousPageCommand;
    
    public bool IsSearching
    {
      get
      {
        return (_IsSearching);
      }
      set
      {
        _IsSearching = value;
        RaisePropertyChanged("IsSearching");
      }
    }
    bool _IsSearching;

    public string SearchText
    {
      get
      {
        return (_SearchText);
      }
      set
      {
        _SearchText = value;
        RaisePropertyChanged("SearchText");
      }
    }
    string _SearchText;

    public ObservableCollection<ImageViewModel> Images
    {
      get
      {
        return (_Images);
      }
      set
      {
        _Images = value;
        RaisePropertyChanged("Images");
      }
    }
    ObservableCollection<ImageViewModel> _Images;

    public int CurrentPage
    {
      get
      {
        return (_CurrentPage);
      }
      set
      {
        _CurrentPage = value;
        RaisePropertyChanged("CurrentPage");
      }
    }
    int _CurrentPage;

    public long TotalPages
    {
      get
      {
        return (_TotalPages);
      }
      set
      {
        _TotalPages = value;
        RaisePropertyChanged("TotalPages");
      }
    }
    long _TotalPages;

    void InitialiseSearchTextSubscription()
    {
      // Attempt is to be notified whenever you change the search text and
      // pick up that change if you stick with it for 750 milliseconds.
      // This is a bit unusual in that I use a Select() call to actually
      // pick up the value of the changed text because INotifyPropertyChanged
      // doesn't deliver you the new value of the property.
      ObservableEx
        .FromPropertyChange(this, "SearchText")
        .Select(p => this.SearchText)
        .Where(text => text.Length > Constants.MinSearchStringLength)        
        .Throttle(TimeSpan.FromMilliseconds(Constants.SearchTextTimeoutMs))
        .ObserveOnDispatcher()
        .Subscribe(OnNewSearch);
    }
    void OnNewSearch(string searchText)
    {
      this.InternalVisualState = MainViewVisualStates.NoResults;
      this.CurrentPage = 1;
      this.OnSearch();
    }
    void OnSearch()
    {
      FlickrSearchUri uri = new FlickrSearchUri()
      {
        SearchText = this.SearchText,
        Page = this.CurrentPage
      };
      this.IsSearching = true;

      try
      {
        // Create a timer that will produce a value every so often (250ms)
        IObservable<long> slowDownTimer = 
          Observable
            .Interval(TimeSpan.FromMilliseconds(Constants.PopulateListBoxDelayMs));

        this.currentSearch =
          ObservableWebRequest       
            .CreateDeferred(uri.Uri) 
            .Select(wr => FlickrPhotoResultsPage.FromWebResponse(wr))
            .ObserveOnDispatcher()
            .Do(UpdateResultsPageDetails)
            .SelectMany(resultsPage => resultsPage.Photos)
            .Zip(slowDownTimer, (result, interval) => result)
            .ObserveOnDispatcher()
            .Subscribe(
              AddImage, EndSearch, () => EndSearch(null));
      }
      catch (Exception ex) // TODO
      {
        EndSearch(ex);
      }
    }
    void EndSearch(Exception ex)
    {
      this.currentSearch = null;
      this.IsSearching = false;

      if (ex == null)
      {
        int removalCount = this.Images.Count - Constants.FlickrSearchPageSize;

        if (removalCount > 0)
        {
          Observable
            .Interval(TimeSpan.FromMilliseconds(Constants.ClearListBoxDelayMs))
            .Take(removalCount)
            .ObserveOnDispatcher()
            .Subscribe(
              i => 
              {
                this.Images.RemoveAt(0);
              });
        }
      }
      else
      {
        // TODO - not very nice to do this down here in the ViewModel
        ErrorWindow.ShowException(ex);
      }
    }
    void UpdateResultsPageDetails(FlickrPhotoResultsPage resultsPage)
    {
      this.TotalPages = resultsPage.Pages;
      this.CurrentPage = resultsPage.PageNo;
      this.InternalVisualState = MainViewVisualStates.Results;
      UpdateCommandAvailability();
    }
    void AddImage(FlickrPhotoResult result)
    {
      this.Images.Insert(this.Images.Count, 
        new ImageViewModel(result.Title, result.ImageUri));
    }
    void InitialseCommands()
    {
      this.nextPageCommand = new SimpleCommand(OnNextPage);
      this.previousPageCommand = new SimpleCommand(OnPrevPage);
      this.cancelCommand = new SimpleCommand(OnCancel);
      UpdateCommandAvailability();
    }
    void UpdateCommandAvailability()
    {
      this.nextPageCommand.SetCanExecute(
        (this.TotalPages > 0) && (this.CurrentPage < this.TotalPages));

      this.previousPageCommand.SetCanExecute(
        (this.TotalPages > 0) && (this.CurrentPage > 1));
    }
    void OnNextPage()
    {
      this.CurrentPage++;
      OnSearch();
    }
    void OnPrevPage()
    {
      this.CurrentPage--;
      OnSearch();
    }
    /// <summary>
    /// No UI for this as yet.
    /// </summary>
    void OnCancel()
    {
      if (this.currentSearch != null)
      {
        // Race here if we ever called it.
        this.currentSearch.Dispose();
        EndSearch(null);
      }
    }
    IDisposable currentSearch;
  }
}